local VERSION = tonumber(("$Revision: 41452 $"):match("%d+"))

CowTip = AceLibrary("AceAddon-2.0"):new("AceDB-2.0", "AceEvent-2.0", "AceConsole-2.0", "FuBarPlugin-2.0", "AceHook-2.1")
local CowTip, self = CowTip, CowTip
CowTip.version = "r" .. VERSION
CowTip.revision = VERSION
CowTip.date = ("$Date: 2007-06-23 17:42:31 -0400 (Sat, 23 Jun 2007) $"):match("%d%d%d%d%-%d%d%-%d%d")

local waterfall = AceLibrary:HasInstance("Waterfall-1.0") and AceLibrary("Waterfall-1.0")

local SharedMedia = AceLibrary("SharedMedia-1.0")
local DogTag = AceLibrary("DogTag-1.0")
local Deformat = AceLibrary("Deformat-2.0")

local _G = _G
local GameTooltip = _G.GameTooltip
local UIParent = _G.UIParent
local GetCursorPosition = _G.GetCursorPosition
local UnitExists = _G.UnitExists
local unpack = _G.unpack
local GetScreenHeight = _G.GetScreenHeight
local GetScreenWidth = _G.GetScreenWidth
local CreateFrame = _G.CreateFrame
local UnitIsFriend = _G.UnitIsFriend
local GetFriendInfo = _G.GetFriendInfo
local GetNumFriends = _G.GetNumFriends
local UnitName = _G.UnitName
local UnitIsUnit = _G.UnitIsUnit
local GetGuildInfo = _G.GetGuildInfo
local UnitIsPlayer = _G.UnitIsPlayer
local InCombatLockdown = _G.InCombatLockdown
local GetWatchedFactionInfo = _G.GetWatchedFactionInfo
local UnitIsDeadOrGhost = _G.UnitIsDeadOrGhost
local UnitIsTapped = _G.UnitIsTapped
local UnitIsTappedByPlayer = _G.UnitIsTappedByPlayer

CowTip:RegisterDB("CowTipDB")
CowTip:RegisterDefaults('profile', {
	scale = 1,
	border = "Blizzard Tooltip",
	unitAnchor = "CURSOR",
	unitOffsetX = 0,
	unitOffsetY = 0,
	unitShow = "always",
	frameAnchor = "BOTTOMRIGHT",
	frameOffsetX = 0,
	frameOffsetY = 0,
	unitFrameShow = "always",
	otherFrameShow = "always",
	font = "Friz Quadrata TT",
	background = "Solid",
	bgColor = {
		guild = {0, 0.15, 0, 1},
		faction = {0.15, 0.15, 0, 1},
		hostilePC = {0.25, 0, 0, 1},
		hostileNPC = {0.15, 0, 0, 1},
		friendlyPC = {0, 0, 0.25, 1},
		friendlyNPC = {0, 0, 0.15, 1},
		other = {0, 0, 0, 1},
		dead = {0.15, 0.15, 0.15, 1},
		tapped = {0.25, 0.25, 0.25, 1},
	},
	borderColor = { 1, 1, 1, 1 },
	padding = 4,
	edgeSize = 4,
	lines = {
		lefts = {
			['*'] = "",
			"[SmartClass?ClassColor][NameRealm]", -- [1]
			"Target:", -- [2]
			"Guild:", -- [3]
			"Guild rank:", -- [4]
			"Level:", -- [5]
			"Race:", -- [6]
			"Class:", -- [7]
			"Faction:", -- [8]
			"Status:", -- [9]
			"Health:", -- [10]
			"[~HasNoMP?Text([TypePower]:)]", -- [11]
			"Experience:", -- [12]
			"Threat:", -- [13]
			"Happiness:", -- [14]
			"Zone:", -- [15]
			"[Rez]", -- [16]
		},
		rights = {
			['*'] = "",
			"", -- [1]
			"[IsUnit#player(mouseovertarget)?Text(<<You>>)] [SmartClass#mouseovertarget?ClassColor#mouseovertarget!NameRealm#mouseovertarget]", -- [2]
			"[Guild:IsEqual([Guild#player])?Green][Guild:Angle]", -- [3]
			"[GuildRank]", -- [4]
			"[Classification] [Level]", -- [5]
			"[SmartRace]", -- [6]
			"[SmartClass:ClassColor]", -- [7]
			"[Faction]", -- [8]
			"[Status]", -- [9]
			"[FractionalHP:Short] [PercentHP:Percent:Paren]", -- [10]
			"[~HasNoMP?Text([FractionalMP:Short] [PercentMP:Percent:Paren])]", -- [11]
			"[CurXP:HideZero?Text([FractionalXP:Short] [PercentXP:Percent:Paren])]", -- [12]
			"[~HasNoThreat?Text([FractionalThreat:Short] [PercentThreat:Percent:Paren])]", -- [13]
			"[IsUnit(pet)?HappyText]", -- [14]
			"[Zone]", -- [15]
			"", -- [16]
		}
	}
})

-- FuBar configuration
CowTip.hasIcon = true
CowTip.hideWithoutStandby = true
CowTip.defaultPosition = "MINIMAP"
CowTip.defaultMinimapPosition = 230
CowTip.hasNoText = true
CowTip.hasNoColor = true
CowTip.cannotDetachTooltip = true
CowTip.independentProfile = true
CowTip.blizzardTooltip = true

local leftGameTooltipStrings = {}
local rightGameTooltipStrings = {}

function CowTip:OnInitialize()
	self:RegisterChatCommand('/CowTip', '/Cow', {
		type = 'execute',
		func = function()
			self:OpenMenu()
		end
	})
	
	GameTooltip:SetOwner(UIParent, "ANCHOR_NONE")
	for i = 1, 50 do
		GameTooltip:AddDoubleLine(' ', ' ')
		leftGameTooltipStrings[i] = _G["GameTooltipTextLeft" .. i]
		rightGameTooltipStrings[i] = _G["GameTooltipTextRight" .. i]
	end
	GameTooltip:Show()
	GameTooltip:Hide()
end

function CowTip:OnEnable()
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	
	self:HookScript(GameTooltip, "OnShow", "GameTooltip_OnShow")
	self:HookScript(GameTooltip, "OnHide", "GameTooltip_OnHide")
	self:HookScript(GameTooltip, "OnTooltipSetUnit", "GameTooltip_OnTooltipSetUnit")
	
	self:Hook("GameTooltip_SetDefaultAnchor", "GameTooltip_SetDefaultAnchor", true)
	
	self:SetScale(nil)
	self:SetFont(nil)
	self:SetTexture(nil, nil, nil)
	
	for i = 1, 30 do
		-- this is done to cache functions initially
		DogTag:Evaluate("mouseover", self.db.profile.lines.lefts[i])
		DogTag:Evaluate("mouseover", self.db.profile.lines.rights[i])
	end
end

function CowTip:OnDisable()
	self:GameTooltip_OnHide(GameTooltip)
	
	self:SetScale(nil)
	self:SetFont(nil)
	self:SetTexture(nil, nil, nil)
end

function CowTip:PLAYER_ENTERING_WORLD()
	-- rescale the tooltip
	self:SetScale(self.db.profile.scale)
end

function CowTip:OnClick()
	if waterfall then
		waterfall:Open("CowTip")
	else
		self:Print("Waterfall-1.0 is required to access the GUI.")
	end
end

local lastNumRegistered = 0
local currentUnit
local currentOwner = nil
local currentSameFaction = false
local linesToAdd = {}
local linesToAddRight = {}
function CowTip:GameTooltip_OnTooltipSetUnit(GameTooltip, ...)
--	self.hooks[this].OnTooltipSetUnit(GameTooltip, ...)
--	do return end
	local myWatchedFaction = GetWatchedFactionInfo()
	currentSameFaction = false
	if myWatchedFaction then
		for i = 1, 10 do
			local left = leftGameTooltipStrings[i]
			if left:GetText() == myWatchedFaction then
				currentSameFaction = true
				break
			end
		end
	end
	
	for i = 3, 30 do
		local left = leftGameTooltipStrings[i]
		local left_text = left:GetText()
		if not left_text or not left:IsShown() then
			break
		elseif Deformat(left_text, DAMAGE_TEMPLATE) then
			for j = i, 30 do
				local left = leftGameTooltipStrings[j]
				linesToAdd[j - i + 1] = left:GetText()
				local right = rightGameTooltipStrings[j]
				linesToAddRight[j - i + 1] = right:IsShown() and right:GetText()
			end
			break
		end
	end
	
	local unit
	if currentOwner ~= UIParent then
		local _
		_, unit = GameTooltip:GetUnit()
	else
		unit = "mouseover"
	end
	
	currentUnit = unit
	
	GameTooltip:ClearLines()
	local db = self.db.profile.lines
	local lefts, rights = db.lefts, db.rights
	local num = 0
	for i = 1, #lefts do
		local left, right = lefts[i], rights[i]
		local left_ev, right_ev
		if left:trim() == "" then
			left = nil
		else
			left_ev = DogTag:Evaluate(unit, left)
			if left_ev:trim() == "" then
				left_ev = nil
			end
		end
		if right:trim() == "" then
			right = nil
		else
			right_ev = DogTag:Evaluate(unit, right)
			if right_ev:trim() == "" then
				right_ev = nil
			end
		end
		if right then
			if right_ev or (left and left_ev and left:find("%[")) then
				num = num + 1
				GameTooltip:AddDoubleLine(left_ev or ' ', right_ev, 1, 1, 1, 1, 1, 1)
				local leftFS = leftGameTooltipStrings[num]
				local rightFS = rightGameTooltipStrings[num]
				DogTag:RegisterFontString(leftFS, GameTooltip, unit, left or ' ')
				DogTag:RegisterFontString(rightFS, GameTooltip, unit, right)
			end
		elseif left then
			if left_ev then
				num = num + 1
				GameTooltip:AddLine(left_ev, 1, 1, 1)
				local leftFS = leftGameTooltipStrings[num]
				DogTag:RegisterFontString(leftFS, GameTooltip, unit, left)
			end
		end
	end
	lastNumRegistered = num
	
	for i,left in ipairs(linesToAdd) do
		local right = linesToAddRight[i]
		if right then
			GameTooltip:AddDoubleLine(left, right, 1, 1, 1, 1, 1, 1)
		else
			GameTooltip:AddLine(left, 1, 1, 1)
		end
		linesToAdd[i] = nil
		linesToAddRight[i] = nil
	end
	
	self:ScheduleEvent("CowTip-ReSetUnit", self.ReSetUnit, 1, self)
end

function CowTip:ReSetUnit()
	if currentUnit then
		GameTooltip:SetUnit(currentUnit)
	end
end

local currentOffsetX, currentOffsetY = 0, 0

local function ReanchorTooltip(this)
	this:ClearAllPoints()
	if currentUnit and not UnitExists(currentUnit) then
		this:Hide()
		return
	end
	local x, y = GetCursorPosition()
	local scale = this:GetEffectiveScale()
	x, y = (x + currentOffsetX) / scale, (y + currentOffsetY) / scale
	this:SetPoint("BOTTOM", UIParent, "BOTTOMLEFT", x, y)
end

function CowTip:GameTooltip_SetDefaultAnchor(this, owner, ...)
	currentOwner = owner
	self.hooks.GameTooltip_SetDefaultAnchor(this, owner, ...)
	
	local db = self.db.profile
	
	local anchor, offsetX, offsetY
	if owner == UIParent then -- world character
		anchor = db.unitAnchor
		offsetX = db.unitOffsetX
		offsetY = db.unitOffsetY
	else -- frame
		anchor = db.frameAnchor
		offsetX = db.frameOffsetX
		offsetY = db.frameOffsetY
	end
	if anchor == "CURSOR" then
		if offsetX == 0 and offsetY == 0 then
			this:SetOwner(owner, "ANCHOR_CURSOR")
		else
			currentOffsetX = offsetX
			currentOffsetY = offsetY
			this:SetScript("OnUpdate", ReanchorTooltip)
		end
	else
		this:SetOwner(owner, "ANCHOR_NONE")
		this:ClearAllPoints()
		this:SetPoint(anchor, UIParent, anchor, offsetX, offsetY)
	end
end

function CowTip:GameTooltip_OnShow(this, ...)
	local show
	if currentOwner == UIParent then
		-- world unit
		show = self.db.profile.unitShow
	elseif currentUnit then
		-- unit frame
		show = self.db.profile.unitFrameShow
	else
		-- non-unit frame
		show = self.db.profile.otherFrameShow
	end
	if show == "noncombat" then
		if InCombatLockdown() then
			this:Hide()
			return
		end
	elseif show == "never" then
		this:Hide()
		return
	end
	self.hooks[this].OnShow(this, ...)
	
--	this:SetScript("OnUpdate", ReanchorUnitTooltip)
	
	self:SetBackgroundColor(nil)
	self:SetBorderColor(nil)
end

function CowTip:GameTooltip_OnHide(this, ...)
	for i = 1, lastNumRegistered do
		DogTag:UnregisterFontString(leftGameTooltipStrings[i])
		DogTag:UnregisterFontString(rightGameTooltipStrings[i])
	end
	lastNumRegistered = 0
	
	currentUnit = nil
	
	this:SetScript("OnUpdate", nil)
	
	self.hooks[this].OnHide(this, ...)
end

function CowTip:SetScale(value)
	if value then
		self.db.profile.scale = value
	else
		value = self.db.profile.scale
	end
	if not self:IsActive() then
		value = 1
	end
	
	GameTooltip:SetScale(value)
end

function CowTip:SetFont(value)
	if value then
		self.db.profile.font = value
	else
		value = self.db.profile.font
	end
	if not self:IsActive() then
		value = "Friz Quadrata TT"
	end
	
	local font = SharedMedia:Fetch('font', value)
	
	local text = leftGameTooltipStrings[1]
	if text:GetFont() == font then
		return
	end
	for i = 1, 50 do
		local left = leftGameTooltipStrings[i]
		local right = rightGameTooltipStrings[i]
		local _, size, style = left:GetFont()
		left:SetFont(font, size, style)
		local _, size, style = right:GetFont()
		right:SetFont(font, size, style)
	end
end

local tmp = {}
local tmp2 = {}
function CowTip:SetTexture(border, background, padding, edgeSize)
	if border then
		self.db.profile.border = border
	else
		border = self.db.profile.border
	end
	if background then
		self.db.profile.background = background
	else
		background = self.db.profile.background
	end
	if padding then
		self.db.profile.padding = padding
	else
		padding = self.db.profile.padding
	end
	if edgeSize then
		self.db.profile.edgeSize = edgeSize
	else
		edgeSize = self.db.profile.edgeSize
	end
	if not self:IsActive() then
		border = "Blizzard Tooltip"
		background = "Blizzard Tooltip"
		padding = 4
		edgeSize = 16
	end
	
	tmp.bgFile = SharedMedia:Fetch('background', background)
	tmp.edgeFile = SharedMedia:Fetch('border', border)
	tmp.tile = false
	tmp.edgeSize = edgeSize
	tmp.insets = tmp2
	tmp2.left = padding
	tmp2.right = padding
	tmp2.top = padding
	tmp2.bottom = padding
	GameTooltip:SetBackdrop(tmp)
	
	self:SetBackgroundColor(nil)
	self:SetBorderColor(nil)
end

function CowTip:SetBorder(value)
	self:SetTexture(value, nil, nil, nil)
end

function CowTip:SetBackground(value)
	self:SetTexture(nil, value, nil, nil)
end

function CowTip:SetPadding(value)
	self:SetTexture(nil, nil, value, nil)
end

function CowTip:SetEdgeSize(value)
	self:SetTexture(nil, nil, nil, value)
end

function CowTip:SetBackgroundColor(given_kind, r, g, b, a)
	local kind = given_kind
	if not kind then
		kind = 'other'
		if currentUnit then
			if UnitIsDeadOrGhost(currentUnit) then
				kind = 'dead'
			elseif UnitIsTapped(currentUnit) and not UnitIsTappedByPlayer(currentUnit) then
				kind = 'tapped'
			elseif currentSameFaction then
				kind = 'faction'
			elseif UnitIsPlayer(currentUnit) then
				if UnitIsFriend("player", currentUnit) then
					local playerGuild = GetGuildInfo("player")
					if playerGuild and playerGuild == GetGuildInfo(currentUnit) or UnitIsUnit("player", currentUnit) then
						kind = 'guild'
					else
						local friend = false
						local name = UnitName(currentUnit)
						for i = 1, GetNumFriends() do
							if GetFriendInfo(i) == name then
								friend = true
								break
							end
						end
						if friend then
							kind = 'guild'
						else
							kind = 'friendlyPC'
						end
					end
				else
					kind = 'hostilePC'
				end
			else
				if UnitIsFriend("player", currentUnit) then
					kind = 'friendlyNPC'
				else
					kind = 'hostileNPC'
				end
			end
		end
	end
	local bgColor = self.db.profile.bgColor[kind]
	if r then
		bgColor[1] = r
		bgColor[2] = g
		bgColor[3] = b
		bgColor[4] = a
	else
		r, g, b, a = unpack(bgColor)
	end
	if given_kind then
		self:SetBackgroundColor(nil)
		return
	end
	if not self:IsActive() then
		r, g, b, a = 0, 0, 0, 1
	end
	GameTooltip:SetBackdropColor(r, g, b, a)
end

function CowTip:SetBorderColor(r, g, b, a)
	local borderColor = self.db.profile.borderColor
	if r then
		borderColor[1] = r
		borderColor[2] = g
		borderColor[3] = b
		borderColor[4] = a
	else
		r, g, b, a = unpack(borderColor)
	end
	if not self:IsActive() then
		r, g, b, a = 1, 1, 1, 1
	end
	GameTooltip:SetBackdropBorderColor(r, g, b, a)
end

local offsetBox

function CowTip:SetUnitAnchor(value)
	if offsetBox then
		offsetBox:GetScript("OnDoubleClick")(offsetBox, "LeftButton")
	end
	self.db.profile.unitAnchor = value
	self.db.profile.unitOffsetX = 0
	self.db.profile.unitOffsetY = 0
end

function CowTip:SetFrameAnchor(value)
	if offsetBox then
		offsetBox:GetScript("OnDoubleClick")(offsetBox, "LeftButton")
	end
	self.db.profile.frameAnchor = value
	self.db.profile.frameOffsetX = 0
	self.db.profile.frameOffsetY = 0
end

local anchorValidate = {
	CURSOR = "Cursor",
	TOPLEFT = "Screen, Top-left",
	TOP = "Screen, Top",
	TOPRIGHT = "Screen, Top-right",
	LEFT = "Screen, Left",
	RIGHT = "Screen, Right",
	BOTTOMLEFT = "Screen, Bottom-left",
	BOTTOM = "Screen, Bottom",
	BOTTOMRIGHT = "Screen, Bottom-right",
	CENTER = "Screen, Middle",
}

local showWhenValidate = {
	never = "Never",
	notcombat = "Out of combat",
	always = "Always",
}

function CowTip:CreateOffsetBox(kind)
	if not offsetBox then
		offsetBox = CreateFrame("Button", "CowTipOffsetBox", UIParent)
		offsetBox:SetWidth(300)
		offsetBox:SetHeight(100)
		offsetBox:SetFrameStrata("TOOLTIP")
		
		local bg = offsetBox:CreateTexture("CowTipOffsetBoxBackground", "BACKGROUND")
		bg:SetTexture(0.7, 0.4, 0) -- orange
		bg:SetAllPoints(offsetBox)
		
		local text = offsetBox:CreateFontString("CowTipOffsetBoxText", "ARTWORK", "GameFontHighlight")
		text:SetPoint("CENTER")
		text:SetText("Click and drag to the position you want.\nDouble-click when finished.")
		
		offsetBox:SetScript("OnDoubleClick", function(this)
			-- cleanup
			this:StopMovingOrSizing()
			this:Hide()
		end)
		
		offsetBox:SetScript("OnDragStart", function(this)
			this:StartMoving()
		end)
		
		offsetBox:SetScript("OnDragStop", function(this)
			local kind = this.kind
			local db = self.db.profile
			local anchor
			if kind == "unit" then
				anchor = db.unitAnchor
			else
				anchor = db.frameAnchor
			end
			local offsetX, offsetY
			if anchor == "BOTTOMLEFT" or anchor == "BOTTOM" or anchor == "BOTTOMRIGHT" then
				offsetY = this:GetBottom()
			elseif anchor == "LEFT" or anchor == "CENTER" or anchor == "RIGHT" then
				local _,y = this:GetCenter()
				offsetY = y - GetScreenHeight()/2
			elseif anchor == "TOPLEFT" or anchor == "TOP" or anchor == "TOPRIGHT" then
				offsetY = this:GetTop() - GetScreenHeight()
			else -- cursor
				offsetY = this:GetBottom() - GetScreenHeight()/2
			end
			if anchor == "TOPLEFT" or anchor == "LEFT" or anchor == "BOTTOMLEFT" then
				offsetX = this:GetLeft()
			elseif anchor == "TOP" or anchor == "CENTER" or anchor == "BOTTOM" or anchor == "CURSOR" then
				offsetX = this:GetCenter() - GetScreenWidth()/2
			else -- bottom
				offsetX = this:GetRight() - GetScreenWidth()
			end
			if kind == "unit" then
				db.unitOffsetX = offsetX
				db.unitOffsetY = offsetY
			else
				db.frameOffsetX = offsetX
				db.frameOffsetY = offsetY
			end
			-- save
			this:StopMovingOrSizing()
		end)
		
		offsetBox:SetMovable(true)
		offsetBox:RegisterForDrag("LeftButton")
		offsetBox:Hide()
	end
	
	offsetBox:Show()
	
	local anchor, offsetX, offsetY
	offsetBox.kind = kind
	local db = self.db.profile
	if kind == "unit" then
		anchor = db.unitAnchor
		offsetX = db.unitOffsetX
		offsetY = db.unitOffsetY
	else
		anchor = db.frameAnchor
		offsetX = db.frameOffsetX
		offsetY = db.frameOffsetY
	end
	
	offsetBox:ClearAllPoints()
	if anchor == "CURSOR" then
		offsetBox:SetPoint("BOTTOM", UIParent, "CENTER", offsetX, offsetY)
	else
		offsetBox:SetPoint(anchor, UIParent, anchor, offsetX, offsetY)
	end
	offsetBox:SetClampedToScreen(true)
end

local lineConfigBox
function CowTip:CreateLineConfigBox()
	if not lineConfigBox then
		lineConfigBox = CreateFrame("Frame", "CowTipLineConfigFrame", UIParent, "DialogBoxFrame")

		lineConfigBox:SetWidth(750)
		lineConfigBox:SetHeight(650)
		lineConfigBox:SetPoint("CENTER")
		lineConfigBox:SetBackdrop({
			bgFile = [[Interface\DialogFrame\UI-DialogBox-Background]], 
		    edgeFile = [[Interface\Tooltips\UI-Tooltip-Border]], 
		    tile = true, tileSize = 16, edgeSize = 16, 
		    insets = { left = 5, right = 5, top = 5, bottom = 5 }
		})
		lineConfigBox:SetBackdropColor(0,0,0,1)
		
		local howto = lineConfigBox:CreateFontString(nil, "ARTWORK", "GameFontHighlightLarge")
		howto:SetPoint("TOP", 0, -16)
		howto:SetText("See http://www.wowace.com/wiki/DogTag-1.0 for details on tag usage.")
		
		lineConfigBox:EnableMouse(true)
		lineConfigBox:SetMovable(true)
		lineConfigBox:RegisterForDrag("LeftButton")
		lineConfigBox:SetClampedToScreen(true)
		lineConfigBox:SetScript("OnDragStart", function(this)
			this:StartMoving()
		end)
		lineConfigBox:SetScript("OnDragStop", function(this)
			this:StopMovingOrSizing()
		end)
		
		local function OnEnterPressed(this)
			local num, side = this.num, this.side
			local text = this:GetText() or ''
			
			text = DogTag:FixCasing(text)
			this:SetText(text)
			
			self.db.profile.lines[side == 'left' and 'lefts' or 'rights'][num] = text
		end
		
		local function OnTabPressed(this)
			local num, side = this.num, this.side
			
			this:ClearFocus()
			
			if not IsShiftKeyDown() then
				if side == "left" then
					this:GetParent()["right" .. num]:SetFocus()
				elseif num < 30 then
					this:GetParent()["left" .. (num+1)]:SetFocus()
				end
			else
				if side == "right" then
					this:GetParent()["left" .. num]:SetFocus()
				elseif num > 1 then
					this:GetParent()["right" .. (num-1)]:SetFocus()
				end
			end
		end
		
		for i = 1, 30 do
			for j = 1, 2 do
				local leftSide = j == 1
				local editBox = CreateFrame("EditBox", "CowTipLineConfigFrameEditBox" .. (leftSide and "Left" or "Right") .. i, lineConfigBox)
				local last
				if leftSide then
					lineConfigBox["left" .. i] = editBox
					last = lineConfigBox["left" .. i-1]
				else
					lineConfigBox["right" .. i] = editBox
					last = lineConfigBox["right" .. i-1]
				end
				editBox:SetFontObject(ChatFontNormal)
				editBox:SetHeight(13)
				if last then
					editBox:SetPoint("TOPLEFT", last, "BOTTOMLEFT", 0, -5)
					editBox:SetPoint("TOPRIGHT", last, "BOTTOMRIGHT", 0, -5)
				else
					if leftSide then
						editBox:SetPoint("TOPLEFT", lineConfigBox, "TOPLEFT", 16, -40)
						editBox:SetPoint("TOPRIGHT", lineConfigBox, "TOP", -8, -40)
					else
						editBox:SetPoint("TOPLEFT", lineConfigBox, "TOP", 8, -40)
						editBox:SetPoint("TOPRIGHT", lineConfigBox, "TOPRIGHT", -16, -40)
					end
				end
				editBox:SetJustifyH("LEFT")
		
				local width = editBox:GetWidth()/2 + 10
				local underline = editBox:CreateTexture(nil, "BACKGROUND")
				editBox.underline = underline
				underline:SetTexture("Interface\\Buttons\\WHITE8X8")
				underline:SetHeight(1)
				underline:SetPoint("TOPLEFT", editBox, "BOTTOMLEFT", 0, -1)
				underline:SetPoint("TOPRIGHT", editBox, "BOTTOMRIGHT", 0, -1)
				editBox:SetAutoFocus(false)
				
				editBox:SetScript("OnEscapePressed", editBox.ClearFocus)
				editBox.num = i
				editBox.side = leftSide and "left" or "right"
				
				editBox:SetScript("OnEnterPressed", OnEnterPressed)
				editBox:SetScript("OnEditFocusLost", OnEnterPressed)
				
				editBox:SetScript("OnTabPressed", OnTabPressed)
			end
		end
	end
	lineConfigBox:Show()
	for i = 1, 30 do
		local left = self.db.profile.lines.lefts[i]
		local right = self.db.profile.lines.rights[i]
		lineConfigBox["left" .. i]:SetText(left)
		lineConfigBox["right" .. i]:SetText(right)
	end
end

CowTip.options = {
	type = 'group',
	handler = CowTip,
	args = {
		lines = {
			name = "Line texts",
			desc = "Here you can describe the text on the assorted lines of the tooltip. These use DogTag-1.0 tags. See http://www.wowace.com/wiki/DogTag-1.0 for details.",
			type = 'execute',
			func = "CreateLineConfigBox",
		},
		size = {
			name = "Size",
			desc = "Set how large the tooltip is",
			type = 'range',
			min = 0.25,
			max = 4,
			step = 0.01,
			bigStep = 0.05,
			get = function()
				return CowTip.db.profile.scale
			end,
			set = "SetScale"
		},
		font = {
			name = "Font",
			desc = "What font face to use.",
			type = 'text',
			validate = SharedMedia:List('font'),
			get = function()
				return self.db.profile.font
			end,
			set = "SetFont",
		},
		bgColor = {
			name = "Background color",
			desc = "Set what color the tooltip's background is.",
			type = 'group',
			pass = true,
			get = function(value)
				return unpack(CowTip.db.profile.bgColor[value])
			end,
			set = "SetBackgroundColor",
			args = {
				guild = {
					name = "Guild and friends",
					desc = "Background color for your guildmates and friends.",
					type = 'color',
					hasAlpha = true,
				},
				faction = {
					name = "Currently watched faction",
					desc = "Background color for the currently watched faction.",
					type = 'color',
					hasAlpha = true,
				},
				hostilePC = {
					name = "Hostile players",
					desc = "Background color for hostile players.",
					type = 'color',
					hasAlpha = true,
				},
				hostileNPC = {
					name = "Hostile non-player characters",
					desc = "Background color for hostile non-player characters.",
					type = 'color',
					hasAlpha = true,
				},
				friendlyPC = {
					name = "Friendly players",
					desc = "Background color for friendly players.",
					type = 'color',
					hasAlpha = true,
				},
				friendlyNPC = {
					name = "Friendly non-player characters",
					desc = "Background color for friendly non-player characters.",
					type = 'color',
					hasAlpha = true,
				},
				dead = {
					name = "Dead",
					desc = "Background color for dead units.",
					type = 'color',
					hasAlpha = true,
				},
				tapped = {
					name = "Tapped",
					desc = "Background color for when a unit is tapped by another.",
					type = 'color',
					hasAlpha = true,
				},
				other = {
					name = "Other",
					desc = "Background color for non-units.",
					type = 'color',
					hasAlpha = true,
				},
			}
		},
		borderColor = {
			name = "Border color",
			desc = "Set what color the tooltip's border is.",
			type = 'color',
			hasAlpha = true,
			get = function()
				return unpack(CowTip.db.profile.borderColor)
			end,
			set = "SetBorderColor"
		},
		border = {
			name = "Border style",
			desc = "Change the border style.",
			type = 'text',
			validate = SharedMedia:List('border'),
			get = function()
				return CowTip.db.profile.border
			end,
			set = "SetBorder",
		},
		background = {
			name = "Background style",
			desc = "Change the background style.\n\nNote: You may need to change the Background color to white to see some of the backgrounds properly.",
			type = 'text',
			validate = SharedMedia:List('background'),
			get = function()
				return CowTip.db.profile.background
			end,
			set = "SetBackground",
		},
		padding = {
			name = "Border padding",
			desc = "The size of the border insets.",
			type = 'range',
			min = 0,
			max = 20,
			step = 1,
			get = function()
				return CowTip.db.profile.padding
			end,
			set = "SetPadding",
		},
		edgeSize = {
			name = "Border size",
			desc = "The size the border takes up.",
			type = 'range',
			min = 0,
			max = 20,
			step = 1,
			get = function()
				return CowTip.db.profile.edgeSize
			end,
			set = "SetEdgeSize",
		},
		unitShow = {
			name = "Show unit tooltip",
			desc = "Show the tooltip for world units if...",
			type = 'text',
			validate = showWhenValidate,
			get = function()
				return CowTip.db.profile.unitShow
			end,
			set = function(value)
				CowTip.db.profile.unitShow = value
			end,
		},
		unitFrameShow = {
			name = "Show unit frame tooltip",
			desc = "Show the tooltip for unit frames if...",
			type = 'text',
			validate = showWhenValidate,
			get = function()
				return CowTip.db.profile.unitFrameShow
			end,
			set = function(value)
				CowTip.db.profile.unitFrameShow = value
			end,
		},
		otherFrameShow = {
			name = "Show non-unit frame tooltip",
			desc = "Show the tooltip for non-unit frames if...",
			type = 'text',
			validate = showWhenValidate,
			get = function()
				return CowTip.db.profile.otherFrameShow
			end,
			set = function(value)
				CowTip.db.profile.otherFrameShow = value
			end,
		},
		ttPosition = {
			name = "Position",
			desc = "Change where the tooltip is located.",
			type = 'group',
			args = {
				unitAnchor = {
					type = 'text',
					name = "Unit anchor",
					desc = "The anchor with which unit mouseover tooltips are located.",
					get = function()
						return CowTip.db.profile.unitAnchor
					end,
					set = "SetUnitAnchor",
					validate = anchorValidate,
					order = 1,
				},
				unitOffset = {
					type = 'execute',
					name = "Set unit anchor offset",
					desc = "Click this to bring up an easy-to-use offset box.",
					func = "CreateOffsetBox",
					passValue = "unit",
					order = 2,
				},
				frameAnchor = {
					type = 'text',
					name = "Frame anchor",
					desc = "The anchor with which tooltips are located when hovering over frames.",
					get = function()
						return CowTip.db.profile.frameAnchor
					end,
					set = "SetFrameAnchor",
					validate = anchorValidate,
					order = 3,
				},
				frameOffset = {
					type = 'execute',
					name = "Set frame anchor offset",
					desc = "Click this to bring up an easy-to-use offset box.",
					func = "CreateOffsetBox",
					passValue = "frame",
					order = 4,
				},
			}
		}
	}
}

if waterfall then
	waterfall:Register("CowTip", 
						"aceOptions", CowTip.options,
						"title", "CowTip Configuration",
						"treeLevels", 5);
end

CowTip.OnMenuRequest = CowTip.options

function CowTip:OnTooltipUpdate()
	GameTooltip:AddLine("CowTip")
	GameTooltip:AddLine(" ")
	--[[
	GameTooltip:AddDoubleLine(LOCKED .. ":", (self.db.profile.settings.locked and YES or NO), 1, 1, 0, 1, 1, 1)
	local text
	if configMode == "solo" then
		text = L["Solo"]
	elseif configMode == "party" then
		text = L["Party"]
	elseif configMode == "raid" then
		text = L["Raid"]
	else
		text = L["Off"]
	end
	GameTooltip:AddDoubleLine(L["Configuration mode"] .. ":", text, 1, 1, 0, 1, 1, 1)
	GameTooltip:AddLine(" ")
	GameTooltip:AddLine(L["|cffffff00Click|r to lock/unlock unit frames."], 0.2, 1, 0.2)
	GameTooltip:AddLine(L["|cffffff00Shift-Click|r to change configuration mode."], 0.2, 1, 0.2)
	]]
	if waterfall then
		GameTooltip:AddLine("|cffffff00Left-Click|r to change settings with a nice GUI configuration.", 0.2, 1, 0.2)
	end
	GameTooltip:AddLine("|cffffff00Right-Click|r to change settings with a drop-down menu.", 0.2, 1, 0.2)
end
